/*
 *  SingIt Lyrics Displayer
 *  Copyright (C) 2000 - 2003 Jan-Marek Glogowski <glogow@stud.fbi.fh-darmstadt.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "macros.h"

#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <ctype.h>

#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>

#include "messagebox.h"
#include "singit_main.h"
#include "singit_creator.h"
#include "singit_about.h"
#include "song.h"
#include "timeout.h"
#include "pixmaps.xpm"


static GtkWidget *singit_creator_win = NULL;
static GtkWidget
		*creator_vbox,
			*creator_handle_box, *creator_toolbar,
				*new_file_button,
				*load_file_button,
				*get_current_song_button,
				*save_file_button,
				*save_file_as_button,
//				*export_as_html_button,
				*timestamp_button,
				*text_edit_button,
				*strip_html_button,
				*strip_timetags_button,
				*tooltips_button,
				*about_button,
				*close_button,
			*creator_scrollwindow, *creator_clist,
			*creator_text_hbox, *creator_text_vscrollbar, *creator_text,
			*creator_status_hbox,
				*creator_status_time_bar,
				*creator_status_mode_bar,
				*creator_status_modified_bar,
			*creator_status_filename_bar;

static GtkWidget *save_question_win = NULL;

typedef struct {

	gboolean   cleanedUp;
	guint      time_context_id;
	guint      mode_context_id;
	guint      modified_context_id;
	guint      filename_context_id;
	gboolean   stamp_used;
	gchar     *time_stamp;
	gchar     *text;
	gchar     *filename;
	gboolean   text_changed;
	gchar     *changed_spaces;
	gboolean   continue_func;

	guint      active_editor;
	guint      active_line;

} CreatorStatus;

CreatorStatus creator_status;

extern SingitConfig *singit_config;
extern GuardedTimeout *timeout_tag;
extern VisPlugin singit_vp;

static void singit_creator_cleanup(void);
static void clear_lyrics(void);
static void new_lyrics_event (GtkWidget *widget, gpointer data);
static void save_lyric_file_event(void);
static gboolean check_user_continue(void);
static void show_save_error_message_box(void);

static void set_filename(gboolean editable)
{
	gchar *statusbarText;

#ifdef DEBUG
	puts("singit_creator.c [set_filename]");
#endif

	if (creator_status.filename) { statusbarText = creator_status.filename; }
	else { statusbarText = ""; }

	if (creator_status.filename_context_id) {
		gtk_statusbar_pop(GTK_STATUSBAR(creator_status_filename_bar), creator_status.filename_context_id);
	}
	else {
		creator_status.filename_context_id =
			gtk_statusbar_get_context_id(GTK_STATUSBAR(creator_status_filename_bar), "filename");
	}

	if (editable) {
		gtk_statusbar_push(GTK_STATUSBAR(creator_status_filename_bar),
			creator_status.mode_context_id, statusbarText);
	}
	else {
		gtk_statusbar_push(GTK_STATUSBAR(creator_status_filename_bar),
			creator_status.mode_context_id,
			_("Lyrics not editable due to limited editor support"));
	}
}

void singit_creator_set_time(guint time)
{
	gchar timeStamp[8];
	static guint last_time = 0;

	if (!singit_creator_win) { return; }
	if (last_time != time / 1000) { last_time = time / 1000; }
	if (time > 5999000) { return; }

	timeStamp[7] = '\0';
	creator_status.stamp_used = FALSE;

	if (creator_status.time_context_id != 0) {
		gtk_statusbar_pop(GTK_STATUSBAR(creator_status_time_bar), creator_status.time_context_id);
	}
	else {
		creator_status.time_context_id =
			gtk_statusbar_get_context_id(GTK_STATUSBAR(creator_status_time_bar), "timestamps");
	}
	sprintf(timeStamp, "[%.2i:%.2i]", last_time / 60, last_time % 60);
	g_free(creator_status.time_stamp);
	creator_status.time_stamp = g_strdup(&timeStamp[0]);
	timeStamp[0] = ' ';
	timeStamp[6] = ' ';
	gtk_statusbar_push(GTK_STATUSBAR(creator_status_time_bar), creator_status.time_context_id, timeStamp);
}

static void message_box_event (GtkWidget *widget, gpointer data)
{
#ifdef DEBUG
	printf("singit_creator.c [message_box_event] : %i\n", MESSAGE_BOX_DIALOG(save_question_win)->result);
#endif

	switch (MESSAGE_BOX_DIALOG(save_question_win)->result) {
	case 1:
		creator_status.continue_func = FALSE;
		message_box_dialog_hide(MESSAGE_BOX_DIALOG(save_question_win));
		break;
	case 2:
		message_box_dialog_hide(MESSAGE_BOX_DIALOG(save_question_win));
	case 3:
		creator_status.continue_func = TRUE;
		break;
	}
}

static void toggle_modified_state (gboolean state)
{
	if (state == creator_status.text_changed) { return; }

#ifdef DEBUG
	puts("singit_creator.c [toggle_modified_state]");
#endif

	if (state) {
		gtk_statusbar_pop(GTK_STATUSBAR(creator_status_modified_bar),
			creator_status.modified_context_id);
	}
	else {
		gtk_statusbar_push(GTK_STATUSBAR(creator_status_modified_bar),
			creator_status.modified_context_id, creator_status.changed_spaces);
		creator_status.stamp_used = FALSE;
	}
	creator_status.text_changed = state;
}

void edit_text_changed_event (GtkEditable *editable, gpointer user_data)
{
	toggle_modified_state(TRUE);
}

gint close_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
#ifdef DEBUG
	puts("singit_creator.c [close_event]");
#endif

	gtk_widget_destroy(singit_creator_win);
	return FALSE;
}

static void clear_text(void)
{
	if (gtk_text_get_length(GTK_TEXT(creator_text)) > 0) {
		gtk_text_set_point(GTK_TEXT(creator_text), 0);
		gtk_text_forward_delete(GTK_TEXT(creator_text), gtk_text_get_length(GTK_TEXT(creator_text)));
	}
}

static gchar* singit_creator_get_clist_text()
{
	gchar *result = NULL, *tmp, *text;
	gint i = 1;

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_get_clist_text]");
#endif

	if (GTK_CLIST(creator_clist)->rows > 0) {
		gtk_clist_get_text(GTK_CLIST(creator_clist), 0, 0, &text);
		result = g_strdup(text);
		while (i < GTK_CLIST(creator_clist)->rows) {
			gtk_clist_get_text(GTK_CLIST(creator_clist), i, 0, &text);
			tmp = result;
			result = g_strconcat(tmp, "\n", text, NULL);
			g_free(tmp);
			i++;
		}
	}
	return result;
}

static void refresh_lyric_text(GtkWidget *widget)
{
	gchar *new_text = NULL;
	gboolean refresh = FALSE;

#ifdef DEBUG
	puts("singit_creator.c [refresh_lyric_text]");
#endif

//	printf("%s\n", gtk_type_name(GTK_OBJECT_TYPE(widget)));
	if (GTK_IS_RADIO_BUTTON(widget)) {
		if (widget == timestamp_button) {
			new_text = singit_creator_get_clist_text();
			refresh = TRUE;
		}
		else if (widget == text_edit_button) {
			new_text =
				g_strdup(gtk_editable_get_chars(GTK_EDITABLE(creator_text), 0, -1));
			refresh = TRUE;
		}
	}
	else if (GTK_IS_CLIST(widget)) {
		new_text = singit_creator_get_clist_text();
		refresh = TRUE;
	}
	else if (GTK_IS_TEXT(widget)) {
		new_text =
			g_strdup(gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1));
		refresh = TRUE;
	}
	if (refresh) {
		if (creator_status.text) { g_free(creator_status.text); }
		creator_status.text = new_text;
	}
}

static void refresh_active_lyric_text()
{
#ifdef DEBUG
	puts("singit_creator.c [refresh_active_lyric_text]");
#endif

	if (GTK_TOGGLE_BUTTON(timestamp_button)->active) {
		refresh_lyric_text(timestamp_button);
	}
	else if (GTK_TOGGLE_BUTTON (text_edit_button)->active) {
		refresh_lyric_text(text_edit_button);
	}
}

static void refresh_widget_text(void)
{
	gchar **lines, *lbr[1][1], *tmp;
	guint i = 0;

#ifdef DEBUG
	puts("singit_creator.c [refresh_widget_text]");
#endif

	tmp = creator_status.text;
	creator_status.text = NULL;
	clear_lyrics();
	creator_status.text = tmp;
	if (!creator_status.text) { return; }
	if (GTK_TOGGLE_BUTTON(timestamp_button)->active) {
		gtk_clist_freeze(GTK_CLIST(creator_clist));
		lines = g_strsplit(creator_status.text, "\n", 0);
		while (lines[i]) {
			lbr[0][0] = lines[i];
			gtk_clist_append(GTK_CLIST(creator_clist), lbr[0]);
			i++;
		}
		g_strfreev(lines);
		gtk_clist_thaw(GTK_CLIST(creator_clist));
	}
	else if (GTK_TOGGLE_BUTTON (text_edit_button)->active) {
		gtk_text_freeze(GTK_TEXT(creator_text));
		gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black,
			NULL, creator_status.text, -1);
		gtk_text_thaw(GTK_TEXT(creator_text));
	}
}

static void singit_creator_set_text()
{
#ifdef DEBUG
	puts("singit_creator.c [singit_creator_set_text]");
#endif

	gtk_text_freeze(GTK_TEXT(creator_text));
	gtk_clist_freeze(GTK_CLIST(creator_clist));
	clear_text();
	if (creator_status.text) { g_free(creator_status.text); }
	creator_status.text = singit_creator_get_clist_text();
	if (creator_status.text  != NULL)
		{ gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black, NULL, creator_status.text, -1); }
	gtk_clist_thaw(GTK_CLIST(creator_clist));
	gtk_text_thaw(GTK_TEXT(creator_text));
}

static void singit_creator_set_lyrics()
{
	gchar **lines, *lbr[1][1];
	guint i = 0;

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_set_lyrics]");
#endif

	if (creator_status.text == NULL) { return; }
	gtk_text_freeze(GTK_TEXT(creator_text));
	gtk_clist_freeze(GTK_CLIST(creator_clist));
	gtk_clist_clear(GTK_CLIST(creator_clist));
	g_free(creator_status.text);
	creator_status.text =
		g_strdup(gtk_editable_get_chars(GTK_EDITABLE(creator_text), 0, -1));
	lines = g_strsplit(creator_status.text, "\n", 0);

	while (lines[i]) {
		lbr[0][0] = lines[i];
		gtk_clist_append(GTK_CLIST(creator_clist), lbr[0]);
		i++;
	}
	g_strfreev(lines);
	gtk_clist_thaw(GTK_CLIST(creator_clist));
	gtk_text_thaw(GTK_TEXT(creator_text));
}

static void strip_html_tags_event (GtkWidget *widget, gpointer data)
{
	gchar *new_text = NULL, *pos_begin, *pos_end, *pos_copy;
	gint len;

	refresh_lyric_text(creator_text);
	if (!creator_status.text) { return; }
	if (!check_user_continue()) { return; }

#ifdef DEBUG
	puts("singit_creator.c [strip_html_tags_event]");
#endif

	new_text = g_new(gchar, strlen(creator_status.text) + 1);
	pos_copy = new_text;
	pos_begin = creator_status.text;
	while (pos_begin) {
		pos_end = strstr(pos_begin, "<");
		if (!pos_end) {
			strcpy(pos_copy, pos_begin);
			pos_begin = NULL;
		}
		else {
			len = pos_end - pos_begin;
			if (len > 0) {
				strncpy (pos_copy, pos_begin, len);
				pos_copy += len;
			}
			pos_begin = strstr(pos_end, ">");
			if (pos_begin) { pos_begin++; }
		}
	}
	g_free(creator_status.text);
	creator_status.text = g_strdup(new_text);
	g_free(new_text);

	refresh_widget_text();
	toggle_modified_state(TRUE);
}

static void strip_time_tags_event (GtkWidget *widget, gpointer data)
{
	gchar *new_text = NULL, *pos_begin, *pos_end, *pos_copy;
	gint len;

	refresh_lyric_text(creator_text);
	if (!creator_status.text) { return; }
	if (!check_user_continue()) { return; }

#ifdef DEBUG
	puts("singit_creator.c [strip_time_tags_event]");
#endif

	new_text = g_new(gchar, strlen(creator_status.text) + 1);
	pos_copy = new_text;
	pos_end = pos_begin = creator_status.text;
	while (pos_begin) {
		pos_end = strstr(pos_end, "[");
		if (!pos_end) {
			strcpy(pos_copy, pos_begin);
			pos_begin = NULL;
		}
		else {
			if ((isdigit(pos_end[1])) && (isdigit(pos_end[2])) && (pos_end[3] == ':') &&
				(isdigit(pos_end[4])) && (isdigit(pos_end[5])) && (pos_end[6] == ']')) {
				len = pos_end - pos_begin;
				strncpy (pos_copy, pos_begin, len);
				pos_copy += len;
				pos_end += 7;
				pos_begin = pos_end;
			}
			else { pos_end++; }
		}
	}
	g_free(creator_status.text);
	creator_status.text = g_strdup(new_text);
	g_free(new_text);

	refresh_widget_text();
	toggle_modified_state(TRUE);
}

static void clear_lyrics(void)
{
#ifdef DEBUG
	puts("singit_creator.c [clear_lyrics]");
#endif

	gtk_text_freeze(GTK_TEXT(creator_text));
	gtk_clist_freeze(GTK_CLIST(creator_clist));
	clear_text();
	gtk_clist_clear(GTK_CLIST(creator_clist));
	if (creator_status.text) {
		g_free(creator_status.text);
		creator_status.text = NULL;
	}
	if (creator_status.filename) {
		g_free(creator_status.filename);
		creator_status.filename = NULL;
	}
	set_filename(TRUE);
	gtk_clist_thaw(GTK_CLIST(creator_clist));
	gtk_text_thaw(GTK_TEXT(creator_text));
	toggle_modified_state(FALSE);
}

static gboolean check_user_continue(void)
{
#ifdef DEBUG
	puts("singit_creator.c [check_user_continue]");
#endif

	if (creator_status.text_changed) {
		message_box_dialog_show(MESSAGE_BOX_DIALOG(save_question_win),
			GTK_WINDOW (save_question_win) );
		if (!creator_status.continue_func) { return FALSE; }
	}

	return TRUE;
}

static void new_lyrics_event (GtkWidget *widget, gpointer data)
{
#ifdef DEBUG
	puts("singit_creator.c [new_lyrics_event]");
#endif

	if (check_user_continue()) { clear_lyrics(); }
}

void lyrics_from_song_event(GtkWidget *widget, gpointer data)
{
	LSong *song = singit_main_get_song();

	if (!song) {
		set_filename(TRUE);
		return;
	}
	if (!song->editable) {
		if (check_user_continue()) {
			clear_lyrics();
			set_filename(FALSE);
		}
		return;
	}
	set_filename(TRUE);

#ifdef DEBUG
	puts("singit_creator.c [lyrics_from_song_event]");
#endif

	if (check_user_continue()) { clear_lyrics(); }
	else { return; }

	gtk_text_freeze(GTK_TEXT(creator_text));
	gtk_clist_freeze(GTK_CLIST(creator_clist));

	if (!l_song_text_found(song)) {
		creator_status.filename = g_strdup(song->fi_lyrics->name);
		set_filename(TRUE);
		gtk_clist_thaw(GTK_CLIST(creator_clist));
		gtk_text_thaw(GTK_TEXT(creator_text));
		return;
	}

	creator_status.text = l_song_create_stream(song, 1);
	creator_status.filename = g_strdup(song->fi_lyrics->name);

	if (GTK_TOGGLE_BUTTON(timestamp_button)->active) {
		gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black, NULL, creator_status.text, -1);
		singit_creator_set_lyrics();
	}
	else if (GTK_TOGGLE_BUTTON (text_edit_button)->active) {
		gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black, NULL, creator_status.text, -1);
	}

	gtk_clist_thaw(GTK_CLIST(creator_clist));
	gtk_text_thaw(GTK_TEXT(creator_text));
	set_filename(TRUE);
	toggle_modified_state(FALSE);
}

void singit_creator_set_mode(gchar *mode)
{
	if (!singit_creator_win) { return; }

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_set_mode]");
#endif

	if (creator_status.mode_context_id != 0) {
		gtk_statusbar_pop(GTK_STATUSBAR(creator_status_mode_bar), creator_status.mode_context_id);
	}
	else {
		creator_status.mode_context_id =
			gtk_statusbar_get_context_id(GTK_STATUSBAR(creator_status_mode_bar), "modes");
	}
	gtk_statusbar_push(GTK_STATUSBAR(creator_status_mode_bar), creator_status.mode_context_id, mode);
}

void editor_radio_event(GtkWidget *widget, gpointer data)
{
	static GtkWidget *last_widget = NULL;
	if ((last_widget == NULL) && (widget == timestamp_button)) {
		last_widget = timestamp_button;
		return;
	}
	if (!GTK_TOGGLE_BUTTON(widget)->active) { return; }
	if (widget != last_widget) { last_widget = widget; }
	else { return; }

	if (GTK_TOGGLE_BUTTON(timestamp_button)->active) {
#ifdef DEBUG
		puts("singit_creator.c [editor_radio_event] : timestamp_button");
#endif
		gtk_widget_set_sensitive(strip_html_button, FALSE);
		gtk_widget_set_sensitive(strip_timetags_button, FALSE);
		gtk_widget_set_usize(GTK_WIDGET(creator_scrollwindow),
			GTK_WIDGET(creator_text_hbox)->allocation.width,
			GTK_WIDGET(creator_text_hbox)->allocation.height);
		refresh_lyric_text(creator_text);
		singit_creator_set_lyrics();
		gtk_widget_hide(creator_text_hbox);
		gtk_widget_show(creator_scrollwindow);
		singit_creator_set_mode(_(" Editor mode: timestamp"));
	}
	else if (GTK_TOGGLE_BUTTON(text_edit_button)->active) {
#ifdef DEBUG
		puts("singit_creator.c [editor_radio_event] : text_edit_button");
#endif
		gtk_widget_set_sensitive(strip_html_button, TRUE);
		gtk_widget_set_sensitive(strip_timetags_button, TRUE);
		gtk_widget_set_usize(GTK_WIDGET(creator_text_hbox),
			GTK_WIDGET(creator_scrollwindow)->allocation.width,
			GTK_WIDGET(creator_scrollwindow)->allocation.height);
		singit_creator_set_text();
		gtk_widget_hide(creator_scrollwindow);
		gtk_widget_show(creator_text_hbox);
		singit_creator_set_mode(_(" Editor mode: text"));
	}
}

static void insert_remove_timestamp_event(GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
	gchar *oldline = NULL, *newline = NULL;
/*	GList *item = NULL;
	gint i;*/

#ifdef DEBUG
	puts("singit_creator.c [insert_remove_timestamp_event]");
#endif

	if (!event) { return; }
	gtk_clist_unselect_row(GTK_CLIST(clist), row, column);

	gtk_clist_get_text(GTK_CLIST(clist), row, column, &oldline);
	creator_status.active_line = row;
	if (event->state & GDK_CONTROL_MASK) {
		if (strlen(oldline) >= 7) {
			if ((oldline[0] == '[') && (isdigit(oldline[1])) && (isdigit(oldline[2])) &&
			(oldline[3] == ':') && (isdigit(oldline[4])) && (isdigit(oldline[5])) && (oldline[6] == ']'))
			{
				if (strlen(oldline) == 7) { newline = g_strdup("\0"); }
				else { newline = g_strdup(&oldline[7]); }
				gtk_clist_set_text(GTK_CLIST(clist), row, column, newline);
			}
		}
	}
	else {
		if (singit_config->addSameTime || !creator_status.stamp_used) {
			creator_status.stamp_used = xmms_remote_is_playing(singit_vp.xmms_session);
			newline = g_strconcat(creator_status.time_stamp, oldline, NULL);
			gtk_clist_set_text(GTK_CLIST(clist), row, column, newline);
			if (((row+1) < GTK_CLIST(clist)->rows) && !(event->state & GDK_SHIFT_MASK)) {
/*				item = GTK_CLIST(creator_clist)->row_list;
				for (i = 1; i < row; i++) { item = g_list_next(item); }
				GTK_CLIST_ROW(item)->state = GTK_STATE_NORMAL;
				GTK_CLIST_ROW(g_list_next(item))->state = GTK_STATE_SELECTED;
				GTK_CLIST(creator_clist)->focus_row++;*/
				gtk_clist_moveto(GTK_CLIST(clist), (row+1), -1, 0.5, 0);
			}
		}
	}
//	g_free(newline);
	toggle_modified_state(TRUE);
}

void tooltips_toggle_event (GtkWidget *widget, gpointer data)
{
#ifdef DEBUG
	puts("singit_creator.c [tooltips_toggle_event]");
#endif

	gtk_toolbar_set_tooltips(GTK_TOOLBAR(data), GTK_TOGGLE_BUTTON(widget)->active );
}

static void ok_button_open_file_selector_event(GtkWidget *widget, GtkWidget *fileSelector)
{
	LSong *song = NULL;

#ifdef DEBUG
	puts("singit_creator.c [ok_button_open_file_selector_event]");
#endif

	song = l_song_new(NULL);
	if (l_song_load_lyrics(song, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileSelector)))) {
		gtk_text_freeze(GTK_TEXT(creator_text));
		gtk_clist_freeze(GTK_CLIST(creator_clist));
		clear_lyrics();
		creator_status.text = l_song_create_stream(song, 1);
		creator_status.filename = g_strdup(song->fi_lyrics->name);
		if (GTK_TOGGLE_BUTTON(timestamp_button)->active) {
			gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black, NULL,
				creator_status.text, -1);
			singit_creator_set_lyrics();
		}
		else if (GTK_TOGGLE_BUTTON (text_edit_button)->active) {
			gtk_text_insert(GTK_TEXT(creator_text), NULL, &creator_text->style->black, NULL,
				creator_status.text, -1);
		}
		set_filename(TRUE);
		gtk_clist_thaw(GTK_CLIST(creator_clist));
		gtk_text_thaw(GTK_TEXT(creator_text));
		toggle_modified_state(FALSE);
		message_box_dialog_hide(MESSAGE_BOX_DIALOG(save_question_win));
	}
	l_song_free(song);
	if (!singit_config->multiSelect)
		{ gtk_widget_destroy(fileSelector); }
}

static void open_lyric_file_event(void)
{
	GtkWidget *fileSelector;
	gchar *file = NULL;

#ifdef DEBUG
	puts("singit_creator.c [open_lyric_file_event]");
#endif

	if (!check_user_continue()) { return; }

	fileSelector = gtk_file_selection_new(_("Lyrics selection"));

	gtk_signal_connect(GTK_OBJECT(fileSelector), "destroy",
		(GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT(fileSelector));

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->ok_button),
		"clicked", (GtkSignalFunc) ok_button_open_file_selector_event, fileSelector);

	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->cancel_button),
		"clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT(fileSelector));

	file = g_strconcat(singit_config->basePath, "*.", singit_config->lyricExtension, NULL);
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileSelector), file);
	g_free(file);

	gtk_widget_show(fileSelector);
}

static void show_save_error_message_box(void)
{
	GtkWidget *message_box = NULL;
	GtkWidget *message_table, *message_label, *message_button_ok;

#ifdef DEBUG
	puts("singit_creator.c [show_save_error_message_box]");
#endif

	message_box = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(message_box), _("Error / Failure"));
	gtk_window_set_modal(GTK_WINDOW(message_box), TRUE);
	gtk_signal_connect_object(GTK_OBJECT(message_box), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(message_box));

	message_table = gtk_table_new (1, 1, TRUE);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(message_box)->vbox), message_table);
	gtk_container_border_width(GTK_CONTAINER(message_table), 7);
	gtk_widget_show(message_table);

	message_label = gtk_label_new (_("File save failed"));
	gtk_table_attach_defaults (GTK_TABLE(message_table), message_label, 0, 1, 0, 1);
	gtk_widget_show(message_label);

	message_button_ok = gtk_button_new_with_label(_("Ok"));
	gtk_signal_connect_object(GTK_OBJECT(message_button_ok), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(message_box));
	GTK_WIDGET_SET_FLAGS(message_button_ok, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(message_box)->action_area), message_button_ok, TRUE, TRUE, 0);
	gtk_widget_show(message_button_ok);

	gtk_widget_show(message_box);
}

static void ok_button_save_as_file_selector_event(GtkWidget *widget, GtkWidget *fileSelector)
{
	LSong *song;
#ifdef DEBUG
	printf("singit_creator.c [ok_button_save_as_file_selector_event] : %s\n", gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileSelector)));
#endif
	song = l_song_new(NULL);
	l_song_build_lyrics(song, creator_status.text);
	if (!l_song_save_lyrics(song, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileSelector))))
		{ show_save_error_message_box(); }
	else {
		g_free(creator_status.filename);
		creator_status.filename = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileSelector)));
		set_filename(TRUE);
		toggle_modified_state(FALSE);
	}
	l_song_free(song);
	gtk_widget_destroy(fileSelector);
}

static void save_as_lyric_file_event(void)
{
	GtkWidget *fileSelector;
	gchar *file = NULL;

#ifdef DEBUG
	puts("singit_creator.c [save_as_lyric_file_event]");
#endif

	refresh_active_lyric_text();

	fileSelector = gtk_file_selection_new(_("Lyric file save"));

	gtk_signal_connect(GTK_OBJECT(fileSelector), "destroy",
		GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fileSelector));

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->ok_button),
		"clicked", GTK_SIGNAL_FUNC(ok_button_save_as_file_selector_event), fileSelector);

	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fileSelector));

	file = g_strconcat(singit_config->basePath, "*.", singit_config->lyricExtension, NULL);
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileSelector), file);
	g_free(file);

	gtk_widget_show(fileSelector);
}

static void save_lyric_file_event(void)
{
	LSong *song = NULL;

#ifdef DEBUG
	puts("singit_creator.c [save_lyric_file_event]");
#endif

	if (!creator_status.filename) { save_as_lyric_file_event(); }
	else {
		refresh_active_lyric_text();
		song = l_song_new(NULL);
		l_song_build_lyrics(song, creator_status.text);
		if (!l_song_save_lyrics(song, creator_status.filename))
			{ save_as_lyric_file_event(); }
		else { toggle_modified_state(FALSE); }
		l_song_free(song);
	}
}

static gboolean singit_creator_main_keypress(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if ((event->keyval == GDK_t) && (event->state & GDK_CONTROL_MASK)) {
		refresh_active_lyric_text();
		if (!GTK_TOGGLE_BUTTON(timestamp_button)->active) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(timestamp_button),TRUE);
			editor_radio_event(timestamp_button, NULL);
		}
		else
			if (!GTK_TOGGLE_BUTTON(text_edit_button)->active) {
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(text_edit_button),TRUE);
				editor_radio_event(text_edit_button, NULL);
			}
	}
	return TRUE;
}

static gboolean singit_creator_clist_keypress(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	gboolean result = TRUE;
	GdkEventButton *button_event = NULL;

	switch(event->keyval) {
		case GDK_Escape:
			singit_creator_hide();
			break;
		case GDK_Insert:
		case GDK_Delete:
			button_event = g_malloc(sizeof(GdkEventButton));
			if (event->keyval == GDK_Insert) { button_event->state = 0; }
			else { button_event->state = (event->state | GDK_CONTROL_MASK); }
			insert_remove_timestamp_event(creator_clist, GTK_CLIST(creator_clist)->focus_row,
				0, button_event, NULL);
			g_free(button_event);
			break;
/*		case GDK_Up:
		case GDK_Down:
		case GDK_Page_Up:
		case GDK_Page_Down:
			gtk_clist_select_row(GTK_CLIST(creator_clist), GTK_CLIST(creator_clist)->focus_row, 0);
			break;*/
		default:
			result = FALSE;
			// printf("%x - %x\n", event->state, event->keyval);
			break;
        }
        return result;
}

static void singit_creator_init(void)
{
#ifdef DEBUG
	puts("singit_creator.c [singit_creator_init]");
#endif

	creator_status.mode_context_id = 0;
	creator_status.time_context_id = 0;
	creator_status.filename_context_id = 0;
	creator_status.text = NULL;
	creator_status.cleanedUp = FALSE;
	creator_status.time_stamp = g_strdup("");
	creator_status.filename = NULL;
	creator_status.stamp_used = FALSE;
	creator_status.text_changed = FALSE;
	creator_status.changed_spaces = NULL;
	creator_status.continue_func = TRUE;
	creator_status.active_editor = 0;
	creator_status.active_line = 0;
}

static void singit_creator_cleanup(void)
{
	if (creator_status.cleanedUp) { return; }

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_cleanup]");
#endif

	creator_status.cleanedUp = TRUE;

	if (creator_status.filename != NULL) {
		g_free(creator_status.filename);
		creator_status.filename = NULL;
	}

	if (creator_status.text != NULL) {
		g_free(creator_status.text);
		creator_status.text = NULL;
	}

	if (creator_status.time_stamp != NULL) {
		g_free(creator_status.time_stamp);
		creator_status.time_stamp = NULL;
	}

	if (creator_status.changed_spaces) {
		g_free(creator_status.changed_spaces);
		creator_status.changed_spaces = NULL;
	}

	creator_status.mode_context_id = 0;
	creator_status.time_context_id = 0;
	creator_status.filename_context_id = 0;
	singit_creator_win = NULL;

	if (save_question_win) {
		gtk_widget_destroy(save_question_win);
	}
	save_question_win = NULL;

	singit_config_save(singit_config);
	if (singit_config_detach(singit_config, TRUE)) { singit_config = NULL; }
	if (guarded_timeout_detach(timeout_tag, TRUE)) { timeout_tag = NULL; }
}

static void singit_creator_close(void)
{
#ifdef DEBUG
	puts("singit_creator.c [singit_creator_close]");
#endif

	save_question_win = NULL;
	singit_creator_win = NULL;
	singit_creator_cleanup();
}

void singit_creator_hide(void)
{
	if (!singit_creator_win) return;

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_hide]");
#endif

	singit_creator_cleanup();
}

void singit_creator_show(void)
{
	GdkPixmap * icon = NULL;
	GdkBitmap * mask = NULL;
	GtkWidget * iconw = NULL;
	gchar *title;

	if (singit_creator_win) { return; }

#ifdef DEBUG
	puts("singit_creator.c [singit_creator_show]");
#endif

#ifdef ENABLE_NLS
	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALE_DIR);
#endif

	singit_creator_init();

	save_question_win = message_box_dialog_new (_("Question"), _("Do you want to save your changes ?"),
		_("  Cancel  "), 100, 50);
	message_box_dialog_prepend_button(MESSAGE_BOX_DIALOG(save_question_win), _(" Continue "), NULL);
	message_box_dialog_prepend_button(MESSAGE_BOX_DIALOG(save_question_win), _("   Save   "),
		GTK_SIGNAL_FUNC(save_lyric_file_event));
	gtk_signal_connect(GTK_OBJECT(save_question_win), "button_pressed",
		GTK_SIGNAL_FUNC(message_box_event), NULL);

	singit_creator_win = gtk_window_new(GTK_WINDOW_DIALOG);

	title = g_strconcat(SINGIT_VERSION_STRING, " - ", _("Lyrics creator"), NULL);
	gtk_window_set_title(GTK_WINDOW(singit_creator_win), title);
	gtk_signal_connect(GTK_OBJECT(singit_creator_win), "key-press-event",
		GTK_SIGNAL_FUNC(singit_creator_main_keypress), NULL);
	g_free(title);
	gtk_window_set_policy(GTK_WINDOW(singit_creator_win), TRUE, TRUE, TRUE);
	gtk_signal_connect(GTK_OBJECT(singit_creator_win), "destroy",
		GTK_SIGNAL_FUNC(singit_creator_close), NULL);
	gtk_container_border_width(GTK_CONTAINER(singit_creator_win), 2);
	gtk_widget_realize(singit_creator_win);

	creator_vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(singit_creator_win), creator_vbox);
	gtk_widget_show(creator_vbox);

	creator_handle_box = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(creator_vbox), creator_handle_box, FALSE, FALSE, 2 );
	gtk_widget_show(creator_handle_box);

	creator_toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
	gtk_container_set_border_width(GTK_CONTAINER(creator_toolbar), 1);
	gtk_toolbar_set_space_size(GTK_TOOLBAR(creator_toolbar), 5);
	gtk_container_add(GTK_CONTAINER(creator_handle_box), creator_toolbar);
	gtk_widget_show(creator_toolbar);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, filenew_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	new_file_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("New lyrics"),
		_("Creates an empty lyrics file"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(new_lyrics_event),
		NULL);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, fileopen_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	load_file_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Load lyrics"),
		_("Loads a lyric file"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(open_lyric_file_event),
		NULL);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, compile_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	get_current_song_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Load song lyrics"),
		_("Loads the lyric file form current song"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(lyrics_from_song_event),
		NULL);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, filesave_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	save_file_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Save lyrics"),
		_("Save the lyric file"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(save_lyric_file_event),
		NULL);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, filesave_all_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	save_file_as_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Save lyrics as"),
		_("Save the lyric file with new filename"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(save_as_lyric_file_event),
		NULL);

/*	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, World_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	export_as_html_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Export html"),
		_("Export lyrics as html file"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(export_as_html_event),
		NULL);*/

	gtk_toolbar_append_space(GTK_TOOLBAR(creator_toolbar));

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, prev_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	timestamp_button = gtk_toolbar_append_element(
		GTK_TOOLBAR(creator_toolbar),
		GTK_TOOLBAR_CHILD_RADIOBUTTON,        /* a type of element */
		NULL,                                 /* pointer to widget */
		_("Timestamp"),                       /* label */
		_("Insert timestamps into lyrics"),   /* tooltip */
		"Private",                            /* tooltip private string */
		iconw,                                /* icon */
		GTK_SIGNAL_FUNC (editor_radio_event), /* signal */
		creator_toolbar);                     /* data for signal */

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, next_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	text_edit_button = gtk_toolbar_append_element(
		GTK_TOOLBAR(creator_toolbar),
		GTK_TOOLBAR_CHILD_RADIOBUTTON,        /* a type of element */
		timestamp_button,                     /* pointer to widget */
		_("Text"),                            /* label */
		_("Edit lyrics text"),                /* tooltip */
		"Private",                            /* tooltip private string */
		iconw,                                /* icon */
		GTK_SIGNAL_FUNC (editor_radio_event), /* signal */
		creator_toolbar);                     /* data for signal */

	gtk_toolbar_append_space(GTK_TOOLBAR(creator_toolbar));

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, World_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	strip_html_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Strip HTML"),
		_("Removes all HTML-tags from the text"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(strip_html_tags_event),
		NULL);
	gtk_widget_set_sensitive(strip_html_button, FALSE);

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, build_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	strip_timetags_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("Strip timetags"),
		_("Removes all timetags from the text"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(strip_time_tags_event),
		NULL);
	gtk_widget_set_sensitive(strip_timetags_button, FALSE);

	gtk_toolbar_append_space(GTK_TOOLBAR(creator_toolbar));

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, search_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	tooltips_button = gtk_toolbar_append_element(
		GTK_TOOLBAR(creator_toolbar),
		GTK_TOOLBAR_CHILD_TOGGLEBUTTON,
		NULL,
		_("Tooltips"),
		_("Toolbar with or without tips"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC (tooltips_toggle_event),
		creator_toolbar);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tooltips_button),TRUE);

	gtk_toolbar_append_space(GTK_TOOLBAR(creator_toolbar));

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, Help_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	about_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),
		_("About"),
		_("About the autor(s) and translator(s) & key bindings"),
		"Private",
		iconw,
		GTK_SIGNAL_FUNC(singit_about_show),
		NULL);

	gtk_toolbar_append_space(GTK_TOOLBAR(creator_toolbar));

	icon = gdk_pixmap_create_from_xpm_d(singit_creator_win->window, &mask, &singit_creator_win->style->white, close_xpm);
	iconw = gtk_pixmap_new(icon, mask);
	close_button = gtk_toolbar_append_item(
		GTK_TOOLBAR(creator_toolbar),  /* our toolbar */
		_("Close"),                    /* button label */
		_("Closes this app"),          /* this button's tooltip */
		"Private",                     /* tooltip private info */
		iconw,                         /* icon widget */
		GTK_SIGNAL_FUNC(close_event),  /* a signal */
		NULL);

	creator_scrollwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(creator_scrollwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(creator_scrollwindow, 300, 160);
	gtk_box_pack_start(GTK_BOX(creator_vbox), creator_scrollwindow, TRUE, TRUE, 0);
	gtk_widget_show (creator_scrollwindow);

	creator_clist = gtk_clist_new(1);
	gtk_signal_connect(GTK_OBJECT(creator_clist), "select_row", GTK_SIGNAL_FUNC(insert_remove_timestamp_event), NULL);
	gtk_signal_connect(GTK_OBJECT(creator_clist), "key-press-event", GTK_SIGNAL_FUNC(singit_creator_clist_keypress), NULL);
	gtk_clist_set_selection_mode(GTK_CLIST(creator_clist), GTK_SELECTION_SINGLE);
	gtk_clist_set_reorderable(GTK_CLIST(creator_clist), FALSE);
	gtk_container_add(GTK_CONTAINER(creator_scrollwindow), creator_clist);
	gtk_clist_set_column_width(GTK_CLIST(creator_clist), 0, 1);
	gtk_widget_show(creator_clist);

	creator_text_hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(creator_vbox), creator_text_hbox);

	creator_text = gtk_text_new (NULL, NULL);
	gtk_box_pack_start(GTK_BOX(creator_text_hbox), creator_text, TRUE, TRUE, 0);
	gtk_text_set_editable(GTK_TEXT(creator_text), TRUE);
	gtk_text_set_line_wrap(GTK_TEXT(creator_text), FALSE);
	gtk_signal_connect(GTK_OBJECT(creator_text), "changed", GTK_SIGNAL_FUNC(edit_text_changed_event), NULL);
	gtk_widget_set_usize(creator_text, 220, 120);
	gtk_widget_show (creator_text);
	gtk_widget_realize(creator_text);

	creator_text_vscrollbar = gtk_vscrollbar_new (GTK_TEXT(creator_text)->vadj);
	gtk_box_pack_start(GTK_BOX(creator_text_hbox), creator_text_vscrollbar, FALSE, FALSE, 0);
	gtk_widget_show(creator_text_vscrollbar);

	/* status bars */
	creator_status_hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(creator_vbox), creator_status_hbox, FALSE, FALSE, 0);
	gtk_widget_show(creator_status_hbox);

	creator_status_time_bar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(creator_status_hbox), creator_status_time_bar, FALSE, FALSE, 0);
	gtk_widget_show(creator_status_time_bar);

	creator_status_mode_bar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(creator_status_hbox), creator_status_mode_bar, TRUE, TRUE, 0);
	gtk_widget_show(creator_status_mode_bar);

	creator_status_modified_bar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(creator_status_hbox), creator_status_modified_bar, FALSE, FALSE, 0);
	creator_status.modified_context_id =
		gtk_statusbar_get_context_id(GTK_STATUSBAR(creator_status_modified_bar), "modified");
	gtk_statusbar_push(GTK_STATUSBAR(creator_status_modified_bar), creator_status.modified_context_id,
		_(" modified "));
	gtk_widget_show(creator_status_modified_bar);
	creator_status.changed_spaces = g_strnfill(strlen(_(" modified ")), ' ');
	gtk_statusbar_push(GTK_STATUSBAR(creator_status_modified_bar), creator_status.modified_context_id,
		creator_status.changed_spaces);

	creator_status_filename_bar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(creator_vbox), creator_status_filename_bar, FALSE, TRUE, 0);
	gtk_widget_show(creator_status_filename_bar);

	gtk_widget_show(singit_creator_win);
	singit_creator_set_time(0);
	singit_creator_set_mode(_(" Editor mode: timestamp"));

	if (!timeout_tag) { timeout_tag = guarded_timeout_new(101, timeout_func, NULL); }
	else { guarded_timeout_attach(timeout_tag); }

	if (singit_config_attach(singit_config)) {
		if ((singit_config->creatorWinPosX < 0) || (singit_config->creatorWinPosY < 0)) {
			gtk_window_set_position(GTK_WINDOW(singit_creator_win), GTK_WIN_POS_MOUSE);
		}
		else {
			gdk_window_move(singit_creator_win->window, singit_config->creatorWinPosX, singit_config->creatorWinPosY);
		}
	}
}
